from typing import Union, Any, List, cast, Dict
import pandas, os, base64, io, gzip
from urllib.parse import quote_plus
from Cryptodome.Util.Padding import unpad
from Cryptodome.Cipher import AES
from .utils import sizeOfFmt, timeago, timestamp
from .helpers import DatasetDesc, getJsonOrPrintError, isAuthenticated, getCurrentBranch
from .secure_storage import getSecureData
from .ux import TableHeader, UserImage, renderTemplate


class TimedDatasetCache:

  def __init__(self, expireSeconds: int):
    self.expireSeconds = expireSeconds
    self.initTime = timestamp()
    self.cache: Dict[str, pandas.DataFrame] = {}

  def _maybeResetCache(self):
    if self.initTime + (self.expireSeconds * 1000) < timestamp():
      self.cache = {}
      self.initTime = timestamp()

  def set(self, key: str, val: pandas.DataFrame):
    self._maybeResetCache()
    self.cache[key] = val

  def get(self, key: str):
    self._maybeResetCache()
    return self.cache.get(key, None)


_datasetCache = TimedDatasetCache(5 * 60)


class DatasetList:

  def __init__(self):
    self._datasets: List[DatasetDesc] = []
    self._iter_current = -1
    resp = getJsonOrPrintError("jupyter/v1/datasets/list")
    if resp and resp.datasets:
      self._datasets = resp.datasets

  def _repr_html_(self):
    if not isAuthenticated():
      return ""
    return self._makeDatasetsHtmlTable()

  def __iter__(self):
    return self

  def __next__(self) -> str:
    self._iter_current += 1
    if self._iter_current < len(self._datasets):
      return self._datasets[self._iter_current].name
    raise StopIteration

  def _makeDatasetsHtmlTable(self):
    if len(self._datasets) == 0:
      return "There are no datasets to show."
    headers = [
        TableHeader("Name", TableHeader.LEFT, isCode=True),
        TableHeader("Owner", TableHeader.CENTER),
        TableHeader("Data Refreshed", TableHeader.RIGHT),
        TableHeader("SQL Updated", TableHeader.RIGHT),
        TableHeader("Rows", TableHeader.RIGHT),
        TableHeader("Bytes", TableHeader.RIGHT),
    ]
    rows: List[List[Union[str, UserImage]]] = []
    for d in self._datasets:
      rows.append([
          d.name,
          UserImage(d.ownerInfo.imageUrl, d.ownerInfo.name),
          timeago(d.recentResultMs) if d.recentResultMs != None else '',
          timeago(d.sqlModifiedAtMs) if d.sqlModifiedAtMs != None else '',
          _fmt_num(d.numRows),
          sizeOfFmt(d.numBytes)
      ])
    return renderTemplate("table", headers=headers, rows=rows)


def list():
  return DatasetList()


def _cacheKey(dsName: str):
  return f"{getCurrentBranch()}/{dsName}"


def get(dsName: str, filter_column: Union[str, None] = None, filter_values: Union[List[Any], None] = None):
  cacheKey = _cacheKey(dsName)
  inDeployment = 'WORKSPACE_ID' in os.environ
  df = _datasetCache.get(cacheKey)
  if df is None or not inDeployment:
    if inDeployment:
      stream = _getFromS3(dsName)
    else:
      stream = _getFromWeb(dsName)
    if stream == None:
      return None
    df = cast(
        pandas.DataFrame,
        pandas.read_csv(  # type: ignore
            stream, sep='|', low_memory=False, na_values=['\\N', '\\\\N']))
    _datasetCache.set(cacheKey, df)

  if df is None:
    raise Exception(f"Unable to fetch dataset '{dsName}'")
  elif filter_column != None and filter_values != None:
    return df[df[filter_column].isin(filter_values)]  # type: ignore
  return df


def _getFromWeb(dsName: str):
  data = getJsonOrPrintError(f'jupyter/v1/datasets/get?dsName={quote_plus(dsName)}')
  if data and data.dsrDownloadInfo:
    return getSecureData(data.dsrDownloadInfo, dsName)
  if not isAuthenticated():
    return None
  return None


def _getFromS3(dsName: str):
  _workspaceId = os.getenv('WORKSPACE_ID')
  _pystateBucket = os.getenv('PYSTATE_BUCKET')
  _pystateKeys = os.getenv('PYSTATE_KEYS')
  if _workspaceId == None or _pystateBucket == None or _pystateKeys == None:
    raise Exception(f"EnvVar Missing: WORKSPACE_ID, PYSTATE_BUCKET, PYSTATE_KEYS")
  import boto3  # type: ignore
  try:
    dsKey = f'{_workspaceId}/datasets/{dsName}/{getCurrentBranch()}'
    s3Obj = boto3.client('s3').get_object(Bucket=_pystateBucket, Key=dsKey)  # type: ignore
    fileKeyEnc = base64.b64decode(s3Obj['Metadata']["x-amz-key"])  # type: ignore
    fileIv = base64.b64decode(s3Obj['Metadata']["x-amz-iv"])  # type: ignore
    for key64 in str(_pystateKeys).split(","):
      cipher = AES.new(base64.b64decode(key64), AES.MODE_ECB)  # type: ignore
      fileKey = unpad(cipher.decrypt(fileKeyEnc), AES.block_size)
      cipher = AES.new(fileKey, AES.MODE_CBC, fileIv)  # type: ignore
      bodyDataEnc = cast(bytes, s3Obj['Body'].read())  # type: ignore
      decState = unpad(cipher.decrypt(bodyDataEnc), AES.block_size)
      return io.BytesIO(gzip.decompress(decState))
  except Exception as err:
    strErr = str(err)
    if 'AccessDenied' in strErr or 'NoSuchKey' in strErr:
      raise Exception('Version not found.')
    else:
      raise err
  return None


def _fmt_num(num: Union[int, Any]):
  if type(num) != int:
    return ""
  return format(num, ",")



import io, os, tempfile, base64, zlib, urllib.request, sys, ssl, requests

from typing import TextIO, Any
from Cryptodome.Cipher import AES
from tqdm import tqdm
import pyaes
from Cryptodome.Cipher import AES
from Cryptodome.Util.Padding import pad
from tqdm.utils import CallbackIOWrapper

from .helpers import ObjectUploadInfo, ResultDownloadInfo, getJsonOrPrintError
from .utils import serialize
from .ux import printTemplate


def getSecureData(dri: ResultDownloadInfo, name: str):
  try:
    if not dri:
      raise Exception("Download info missing from API response.")
    _storeDatasetResultIfMissing(name, dri.id, dri.signedDataUrl)
    rawDecryptedData = _decryptUnzipFile(dri.id, dri.key64, dri.iv64)
    return io.BytesIO(rawDecryptedData)
  except Exception as err:
    printTemplate("error", None, errorText=f'Unable to fetch data. ({str(err)})')
    if dri:
      _clearTmpFile(dri.id)
    return None


def _tmpFilepath(dId: str):
  mbTempDir = os.path.join(tempfile.gettempdir(), 'modelbit')
  if not os.path.exists(mbTempDir):
    os.makedirs(mbTempDir)
  return os.path.join(mbTempDir, dId)


def _storeDatasetResultIfMissing(dName: str, dId: str, url: str):
  filepath = _tmpFilepath(dId)
  if os.path.exists(filepath):
    return

  class DownloadProgressBar(tqdm):  # From https://github.com/tqdm/tqdm#hooks-and-callbacks

    def update_to(self, b: int = 1, bsize: int = 1, tsize: None = None):
      if tsize is not None:
        self.total = tsize
      self.update(b * bsize - self.n)  # type: ignore

  outputStream: TextIO = sys.stdout
  if os.getenv('MB_TXT_MODE'):
    outputStream = io.StringIO()
  with DownloadProgressBar(unit='B',
                           unit_scale=True,
                           miniters=1,
                           desc=f'Downloading "{dName}"',
                           file=outputStream) as t:
    default_context = ssl._create_default_https_context  # type: ignore
    try:
      urllib.request.urlretrieve(url, filename=filepath, reporthook=t.update_to)  # type: ignore
    except:
      # In case client has local SSL cert issues: pull down encrypted file without cert checking
      _clearTmpFile(dId)
      ssl._create_default_https_context = ssl._create_unverified_context  # type: ignore
      urllib.request.urlretrieve(url, filename=filepath, reporthook=t.update_to)  # type: ignore
    finally:
      ssl._create_default_https_context = default_context  # type: ignore


def _clearTmpFile(dId: str):
  filepath = _tmpFilepath(dId)
  if os.path.exists(filepath):
    os.remove(filepath)


def _decryptUnzipFile(dId: str, key64: str, iv64: str):
  filepath = _tmpFilepath(dId)
  if not os.path.exists(filepath):
    printTemplate("error", None, errorText=f"Unable to find data at {filepath}")

  fileIn = open(filepath, 'rb')
  raw = fileIn.read()
  fileIn.close()

  try:
    cipher = AES.new(base64.b64decode(key64), AES.MODE_CBC, iv=base64.b64decode(iv64))  # type: ignore
    return zlib.decompress(cipher.decrypt(raw), zlib.MAX_WBITS | 32)
  except Exception:
    # Fallback needed to support: Windows 11 on Mac M1 in Parallels
    printTemplate("error",
                  None,
                  errorText=f"Unable to use native decryption. Falling back to pure-Python decryption.")
    mode = pyaes.AESModeOfOperationCBC(base64.b64decode(key64), iv=base64.b64decode(iv64))  # type: ignore
    outStream = io.BytesIO()
    pyaes.decrypt_stream(mode, io.BytesIO(raw), outStream)  # type: ignore
    outStream.seek(0)
    return zlib.decompress(outStream.read(), zlib.MAX_WBITS | 32)


def uploadRuntimeObject(obj: Any, desc: str):
  (data, contentHash) = serialize(obj)
  resp = getJsonOrPrintError("jupyter/v1/runtimes/get_object_upload_url", {
      "contentHash": contentHash,
  })
  if resp and resp.objectUploadInfo:
    putSecureData(resp.objectUploadInfo, data, desc)
    return contentHash
  raise Exception(f"Unable to upload object.")


def putSecureData(oui: ObjectUploadInfo, data: bytes, desc: str):
  if oui.objectExists:
    return
  cipher = AES.new(  # type: ignore
      mode=AES.MODE_CBC, key=base64.b64decode(oui.key64), iv=base64.b64decode(oui.iv64))
  body = cipher.encrypt(pad(data, AES.block_size))
  outputStream: TextIO = sys.stdout
  if os.getenv('MB_TXT_MODE'):
    outputStream = io.StringIO()
  with io.BytesIO(body) as b:
    with tqdm(total=len(data),
              unit='B',
              unit_scale=True,
              miniters=1,
              desc=f"Uploading '{desc}'",
              file=outputStream) as t:
      wrapped_data = CallbackIOWrapper(t.update, b, "read")
      requests.put(oui.signedDataUrl, data=wrapped_data)  # type: ignore



__version__ = "0.12.3"
__author__ = 'Modelbit'

import os, sys, yaml
from typing import cast, Union, Callable, Any, Dict, List

# aliasing since some of these overlap with functions we want to expose to users
from . import datasets as m_datasets
from . import warehouses as m_warehouses
from . import deployments as m_deployments
from . import model as m_model
from . import runtime as m_runtime
from . import utils as m_utils
from . import helpers as m_helpers
from . import model_wrappers as m_model_wrappers

m_helpers.pkgVersion = __version__


# Nicer UX for customers: from modelbit import Deployment
class Deployment(m_runtime.Deployment):
  ...


class Model(m_model.Model):
  ...


def __str__():
  return "Modelbit Client"


def _repr_html_():  # type: ignore
  return __str__()


datasets = m_datasets.list

get_dataset = m_datasets.get

warehouses = m_warehouses.list

deployments = m_deployments.list

models = m_model.list


def save_model(modelObj: Any,
               name: Union[str, None] = None,
               properties: Union[Dict[str, Any], None] = None,
               helpers: Union[Dict[str, Any], None] = None):
  if type(modelObj) in [Model, m_model.Model]:
    modelObjM = cast(Model, modelObj)
    return modelObjM.save()
  return Model(modelObj, name=name, properties=properties, helpers=helpers).save()


load_model = m_model.get


def deploy(deployableObj: Union[Callable[..., Any], m_runtime.Deployment],
           name: Union[str, None] = None,
           python_version: Union[str, None] = None,
           python_packages: Union[List[str], None] = None,
           system_packages: Union[List[str], None] = None):
  if _objIsDeployment(deployableObj):
    deployableObj = cast(Deployment, deployableObj)
    return deployableObj.deploy()
  elif callable(deployableObj):
    dep = Deployment(
        name=name,
        deploy_function=deployableObj,
        python_version=python_version,
        python_packages=python_packages,
        system_packages=system_packages,
    )
    return dep.deploy()
  elif hasattr(deployableObj, "__module__") and "sklearn" in deployableObj.__module__ and hasattr(
      deployableObj, "predict"):
    return m_model_wrappers.SklearnPredictor(deployableObj,
                                             name=name,
                                             python_version=python_version,
                                             python_packages=python_packages,
                                             system_packages=system_packages).makeDeployment().deploy()
  else:
    raise Exception("First argument must be a function or Deployment object.")


def login(region: Union[str, None] = None):
  m_helpers.performLogin(refreshAuth=True, region=region)
  return sys.modules['modelbit']


def switch_branch(branch: str):
  m_helpers.setCurrentBranch(branch)


isAuthenticated = m_helpers.isAuthenticated


def load_value(name: str):
  # if "snowparkZip" in os.environ:
  #   zipPath = [p for p in sys.path if p.endswith(os.environ["snowparkZip"])][0]
  #   importer = zipimport.zipimporter(zipPath)
  #   # Typing thinks the response is a string, but we get bytes
  #   val64 = cast(bytes, importer.get_data(f"{name}.pkl")).decode()
  #   return m_utils.unpickleObj(val64)
  extractPath = os.environ['MB_EXTRACT_PATH']
  objPath = os.environ['MB_RUNTIME_OBJ_DIR']
  if not extractPath or not objPath:
    raise Exception("Missing extractPath/objPath")
  with open(f"{extractPath}/metadata.yaml", "r") as f:
    yamlData = cast(Dict[str, Any], yaml.load(f, Loader=yaml.SafeLoader))  # type: ignore
  data: Dict[str, Dict[str, str]] = yamlData["data"]
  contentHash = data[name]["contentHash"]
  with open(f"{objPath}/{contentHash}.pkl.gz", "rb") as f:
    return m_utils.deserialize(contentHash, f.read)


def _objIsDeployment(obj: Any):
  try:
    if type(obj) in [Deployment, m_runtime.Deployment]:
      return True
    # catch modelbit._reload() class differences
    if obj.__class__.__name__ in ['Deployment']:
      return True
  except:
    return False
  return False


# def _reload():  # type: ignore
#   import importlib
#   importlib.reload(modelbit_core)
#   importlib.reload(datasets)
#   importlib.reload(warehouses)
#   importlib.reload(runtime)
#   importlib.reload(model)
#   importlib.reload(deployments)
#   importlib.reload(secure_storage)
#   importlib.reload(utils)
#   importlib.reload(ux)
#   importlib.reload(helpers)
#   importlib.reload(model_wrappers)
#   importlib.reload(importlib.import_module("modelbit"))
#   print("All modules reloaded, except session.")



from typing import Union, Any, cast, Dict, Callable
from datetime import datetime
import os, re, pickle, codecs, gzip, hashlib
import time

_deserializeCache: Dict[str, Any] = {}


# From https://stackoverflow.com/questions/1094841/get-human-readable-version-of-file-size
def sizeOfFmt(num: Union[int, Any]):
  if type(num) != int:
    return ""
  numLeft: float = num
  for unit in ["Bytes", "KB", "MB", "GB", "TB", "PB", "EB", "ZB"]:
    if abs(numLeft) < 1000.0:
      return f"{numLeft:3.0f} {unit}"
    numLeft /= 1000.0
  return f"{numLeft:.1f} YB"


def pandasTypeToPythonType(pandasType: str):
  if pandasType in ['float32', 'float64']:
    return 'float'
  if pandasType in ['int32', 'int64']:
    return 'int'
  if pandasType == 'bool':
    return 'bool'
  return 'Any'


def simplifyArgName(argName: str):
  scrubbed = re.sub("\\W+", "_", argName.lower())
  scrubbed = re.sub('^(\\d+)', "c\\1", scrubbed)
  if scrubbed.endswith("_"):
    scrubbed = scrubbed[:-1]
  return scrubbed


def unindent(source: str) -> str:
  leadingWhitespaces = len(source) - len(source.lstrip())
  if leadingWhitespaces == 0:
    return source
  newLines = [line[leadingWhitespaces:] for line in source.split("\n")]
  return "\n".join(newLines)


def timeago(pastDateMs: int):
  nowMs = time.time() * 1000
  options = [
      {
          "name": "second",
          "divide": 1000
      },
      {
          "name": "minute",
          "divide": 60
      },
      {
          "name": "hour",
          "divide": 60
      },
      {
          "name": "day",
          "divide": 24
      },
      {
          "name": "month",
          "divide": 30.5
      },
  ]
  currentDiff = nowMs - pastDateMs
  if currentDiff < 0:
    raise Exception("The future is NYI")
  resp = "Just now"
  for opt in options:
    currentDiff = round(currentDiff / cast(Union[float, int], opt["divide"]))
    if currentDiff <= 0:
      return resp
    pluralS = ""
    if currentDiff != 1:
      pluralS = "s"
    resp = f"{currentDiff} {opt['name']}{pluralS} ago"
  return resp


def deprecated_unpickleObj(str64: str):
  return pickle.loads(codecs.decode(str64.encode(), "base64"))


def deprecated_pickleObj(obj: Any):
  return codecs.encode(pickle.dumps(obj), "base64").decode()


def deserialize(contentHash: str, reader: Callable[..., Any]):
  if contentHash not in _deserializeCache:
    _deserializeCache[contentHash] = pickle.loads(gzip.decompress(reader()))
  return _deserializeCache[contentHash]


def serialize(obj: Any):
  pklData = pickle.dumps(obj)
  zippedPickle = gzip.compress(pklData, compresslevel=1)
  contentHash = hashlib.md5(pklData).hexdigest()
  return (zippedPickle, contentHash)


def timestamp():
  return int(datetime.timestamp(datetime.now()) * 1000)


def getEnvOrDefault(key: str, defaultVal: str) -> str:
  osVal = os.getenv(key)
  if type(osVal) == str:
    return str(osVal)
  else:
    return defaultVal



from typing import List

from .helpers import GenericWarehouse, getJsonOrPrintError, isAuthenticated
from .utils import timeago
from .ux import TableHeader, renderTemplate


class WarehousesList:

  def __init__(self):
    self._warehouses: List[GenericWarehouse] = []
    resp = getJsonOrPrintError("jupyter/v1/warehouses/list")
    if resp and resp.warehouses:
      self._warehouses = resp.warehouses

  def _repr_html_(self):
    if not isAuthenticated():
      return ""
    return self._makeWarehousesHtmlTable()

  def _makeWarehousesHtmlTable(self):
    if len(self._warehouses) == 0:
      return ""
    headers = [
        TableHeader("Name", TableHeader.LEFT, isCode=True),
        TableHeader("Type", TableHeader.LEFT),
        TableHeader("Connected", TableHeader.LEFT),
        TableHeader("Deploy Status", TableHeader.LEFT),
    ]
    rows: List[List[str]] = []
    for w in self._warehouses:
      connectedAgo = timeago(w.createdAtMs)
      rows.append([w.displayName, str(w.type), connectedAgo, w.deployStatusPretty])
    return renderTemplate("table", headers=headers, rows=rows)


def list():
  return WarehousesList()


from typing import List, Union, Tuple, Dict
import os, sys, json

ALLOWED_PY_VERSIONS = ['3.6', '3.7', '3.8', '3.9', '3.10']


def listInstalledPackages():
  return [f'{p["name"]}=={p["version"]}' for p in getPipList()]


# Returns List[(desiredPackage, installedPackage|None)]
def listMissingPackagesFromPipList(
    deploymentPythonPackages: Union[List[str], None]) -> List[Tuple[str, Union[str, None]]]:
  missingPackages: List[Tuple[str, Union[str, None]]] = []

  if deploymentPythonPackages is None or len(deploymentPythonPackages) == 0:
    return missingPackages

  installedPackages = listInstalledPackages()
  for dpp in deploymentPythonPackages:
    if dpp not in installedPackages:
      similarPackage: Union[str, None] = None
      dppNoVersion = dpp.split("=")[0].lower()
      for ip in installedPackages:
        if ip.startswith(dppNoVersion):
          similarPackage = ip
      missingPackages.append((dpp, similarPackage))

  return missingPackages


def getInstalledPythonVersion():
  installedVer = f"{sys.version_info.major}.{sys.version_info.minor}"
  return installedVer


def packagesToIgnoreFromImportCheck(deploymentPythonPackages: Union[List[str], None]) -> List[str]:
  ignorablePackages: List[str] = []
  if deploymentPythonPackages is None:
    return ignorablePackages

  missingPackages = listMissingPackagesFromPipList(deploymentPythonPackages)
  for mp in missingPackages:
    if mp[1] is not None:
      ignorablePackages.append(mp[1].split("=")[0])

  return ignorablePackages


# Returns List[(importedModule, pipPackageInstalled)]
def listMissingPackagesFromImports(importedModules: Union[List[str], None],
                                   deploymentPythonPackages: Union[List[str], None]) -> List[Tuple[str, str]]:
  missingPackages: List[Tuple[str, str]] = []
  ignoreablePackages = packagesToIgnoreFromImportCheck(deploymentPythonPackages)
  if importedModules is None:
    return missingPackages
  if deploymentPythonPackages is None:
    deploymentPythonPackages = []

  installedModules = listInstalledPackagesByModule()
  for im in importedModules:
    baseModule = im.split(".")[0]
    if baseModule not in installedModules:
      continue  # module is likely a system module, e.g. json
    pipInstall = installedModules[baseModule]
    pipPackage = pipInstall.split("=")[0]
    if pipInstall not in deploymentPythonPackages and pipPackage not in ignoreablePackages:
      missingPackages.append((im, pipInstall))

  return missingPackages


def getPipInstallAndModuleFromDistInfo(distInfoPath: str) -> Dict[str, str]:
  try:
    tPath = os.path.join(distInfoPath, "top_level.txt")
    moduleNames: List[str] = []
    if not os.path.exists(tPath):
      return {}
    with open(tPath) as f:
      moduleNames = f.read().strip().split("\n")

    mPath = os.path.join(distInfoPath, "METADATA")
    if not os.path.exists(mPath):
      return {}

    pipName = None
    pipVersion = None
    with open(mPath) as f:
      metadata = f.read().split("\n")
      for mLine in metadata:
        if mLine.startswith("Name: "):
          pipName = mLine.split(":")[1].strip()
        if mLine.startswith("Version: "):
          pipVersion = mLine.split(":")[1].strip()
        if pipName is not None and pipVersion is not None:
          break

    if pipName is None or pipVersion is None:
      return {}

    modulesToPipVersions: Dict[str, str] = {}
    for moduleName in moduleNames:
      modulesToPipVersions[moduleName] = f"{pipName}=={pipVersion}"
    return modulesToPipVersions
  except Exception as err:
    print(f"Warning, unable to check module '{distInfoPath}': {err}")
    return {}


def listInstalledPackagesByModule():
  packages = getPipList()
  installPaths: Dict[str, int] = {}
  for package in packages:
    installPaths[package["location"]] = 1

  modulesToPipVersions: Dict[str, str] = {}
  for installPath in installPaths.keys():
    for fileOrDir in os.listdir(installPath):
      if fileOrDir.endswith("dist-info"):
        dPath = os.path.join(installPath, fileOrDir)
        modulesToPipVersions.update(getPipInstallAndModuleFromDistInfo(dPath))

  return modulesToPipVersions


def getPipList():
  return json.loads(os.popen("pip list -v --format json --disable-pip-version-check").read().strip())



from typing import Any, Union, List
import inspect, re, string
from .runtime import Deployment


def make_json_friendly(value: Any):
  import numpy
  if isinstance(value, numpy.ndarray):
    return value.tolist()
  if isinstance(value, numpy.integer):
    return int(value)  # type: ignore
  if isinstance(value, numpy.floating):
    return float(value)  # type: ignore
  return value


class SklearnPredictor:

  def __init__(self,
               skpredictor: Any,
               name: Union[str, None] = None,
               python_version: Union[str, None] = None,
               python_packages: Union[List[str], None] = None,
               system_packages: Union[List[str], None] = None):
    self.skpredictor = skpredictor
    self.python_version = python_version
    self.python_packages = python_packages
    self.system_packages = system_packages
    if name:
      self.name = name
    else:
      self.name = self.guessModelName()

  def guessModelName(self):
    try:
      codeContexts = [f.code_context for f in inspect.stack()]
      for ccList in codeContexts:
        if not ccList:
          continue
        for cc in ccList:
          captures = re.search(r"\.(deploy|train)\(([^\s,)]+)", cc)
          if captures:
            return captures.group(2)
    except Exception as _:
      pass
    return None

  def guessNumArgs(self):
    for i in range(1, 25):
      try:
        args = [j for j in range(i)]
        self.skpredictor.predict([args])  # type: ignore
        return i
      except Exception:
        pass
    return None

  def makeDeployment(self):
    skpredictor = self.skpredictor

    varName = "skpredictor"
    if self.name:
      varName = self.name
    globals()[varName] = skpredictor  # put the same value to globals so it acts more like a notebook cell

    guessedArgCount = self.guessNumArgs()
    if guessedArgCount:
      letters = list(string.ascii_lowercase)
      argNames = letters[0:guessedArgCount]
      argsWithTypes = ", ".join([f"{ltr}: float" for ltr in argNames])
      funcSource = "\n".join([
          f"def predict({argsWithTypes}) -> float:",
          f"  return make_json_friendly({varName}.predict([[{', '.join(argNames)}]])[0])", f""
      ])
    else:
      funcSource = "\n".join(
          [f"def predict(*args: Any):", f"  return make_json_friendly({varName}.predict([args])[0])", f""])

    exec(funcSource)
    deploy_function = locals()["predict"]

    return Deployment(deploy_function=deploy_function,
                      source_override=funcSource,
                      python_version=self.python_version,
                      python_packages=self.python_packages,
                      system_packages=self.system_packages,
                      name=self.name)



from typing import Union, List, Dict, Any, Callable, Tuple
import inspect, re, json

from .environment import ALLOWED_PY_VERSIONS, getInstalledPythonVersion
from .utils import unindent
from .helpers import DeploymentTestDef, DeploymentTestError, RuntimeFile, RuntimePythonProps, RuntimeType, getJson, getJsonOrPrintError, getMissingPackageWarningsFromEnvironment, getMissingPackageWarningsFromImportedModules, isAuthenticated, getDefaultPythonVersion, getDefaultPythonPackages, getDefaultSystemPackages
from .ux import COLORS, DifferentPythonVerWarning, GenericError, GenericTip, WarningErrorTip, makeCssStyle, TableHeader, printTemplate, renderTemplate
from .secure_storage import uploadRuntimeObject


class RuntimeStatusNotes:

  def __init__(self, tips: List[WarningErrorTip], warnings: List[WarningErrorTip],
               errors: List[WarningErrorTip]):
    self.tips = tips
    self.warnings = warnings
    self.errors = errors
    self.deployable = len(errors) == 0

  def statusMsg(self):
    if self.deployable:
      return 'Ready'
    return 'Not Ready'

  def statusStyle(self):
    if self.deployable:
      return "color:green; font-weight: bold;"
    return "color:gray; font-weight: bold;"


class NamespaceCollection:

  def __init__(self):
    self.functions: Dict[str, str] = {}
    self.vars: Dict[str, Any] = {}
    self.imports: Dict[str, str] = {}
    self.froms: Dict[str, str] = {"*": "typing"}
    self.allModules: List[str] = []


class Runtime:
  _runtimeTypeName = "runtime"

  MAX_REQUIREMENTS_TXT = 20_000
  LAMBDA_RAM_MAX_MB = 10_240
  LAMBDA_RAM_MIN_MB = 128

  def __init__(self,
               rtType: RuntimeType,
               name: Union[str, None] = None,
               main_function: Union[Callable[..., Any], None] = None,
               python_version: Union[str, None] = None,
               requirements_txt_filepath: Union[str, None] = None,
               requirements_txt_contents: Union[List[str], None] = None,
               python_packages: Union[List[str], None] = None,
               system_packages: Union[List[str], None] = None,
               docker_run: Union[str, None] = None,
               source_override: Union[str, None] = None):
    self._pythonPackages: Union[List[str], None] = None
    self._systemPackages: Union[List[str], None] = None
    self._dockerRun: Union[str, None] = None
    self._deployName: Union[str, None] = None
    self._deployFunc: Union[Callable[..., Any], None] = None
    self.rtType: RuntimeType = rtType
    self._sourceOverride = source_override
    self._testOutputs: List[Tuple[bool, DeploymentTestDef, Dict[str, str]]] = []

    self._pythonVersion = getDefaultPythonVersion()
    self.set_python_packages(getDefaultPythonPackages())
    self.set_system_packages(getDefaultSystemPackages())

    if name:
      self.set_name(name)
    if main_function:
      self._set_main_function(main_function)
    if python_version:
      self.set_python_version(python_version)
    if requirements_txt_filepath:
      self.set_requirements_txt(filepath=requirements_txt_filepath)
    if requirements_txt_contents:
      self.set_requirements_txt(contents=requirements_txt_contents)
    if python_packages is not None:
      self.set_python_packages(python_packages)
    if system_packages is not None:
      self.set_system_packages(system_packages)
    if docker_run:
      self.set_docker_run(docker_run)
    self._prepEnvironment()

  def _repr_html_(self):
    return self._describeHtml()

  def set_name(self, name: str):
    if not re.match('^[a-zA-Z0-9_]+$', name):
      raise Exception("Names should be alphanumeric with underscores.")
    self._deployName = name
    return self

  def set_python_version(self, version: str):
    if version not in ALLOWED_PY_VERSIONS:
      return self._selfError(f'Python version should be one of {ALLOWED_PY_VERSIONS}.')
    self._pythonVersion = version
    return self

  def set_requirements_txt(self, filepath: Union[str, None] = None, contents: Union[List[str], None] = None):
    # if filepath == None and contents == None:
    # return self._filepicker_requirements_txt()
    lines: List[str] = []
    if filepath != None and type(filepath) == str:
      f = open(filepath, "r")
      lines = [n.strip() for n in f.readlines()]
      return self.set_python_packages(lines)
    elif contents != None:
      return self.set_python_packages(contents)
    return self

  def set_python_packages(self, packages: Union[List[str], None]):
    if packages is None:
      self._pythonPackages = None
      return
    if type(packages) != list:
      raise Exception("The python_packages parameter must be a list of strings.")
    for pkg in packages:
      if type(pkg) != str:
        raise Exception("The python_packages parameters must be a list of strings.")
      if "\n" in pkg or "\r" in pkg:
        raise Exception("The python_packages parameters cannot contain newlines")
      if "==" not in pkg:
        raise Exception(
            f"The python_packages parameter '{pkg}' is formatted incorrectly. It should look like 'package-name==X.Y.Z'"
        )
    self._pythonPackages = packages

  def set_system_packages(self, packages: Union[List[str], None]):
    if packages is None:
      self._systemPackages = None
      return
    if type(packages) != list:
      raise Exception("The system_packages parameter must be a list of strings.")
    for pkg in packages:
      if type(pkg) != str:
        raise Exception("The system_packages parameters must be a list of strings.")
      if "\n" in pkg or "\r" in pkg:
        raise Exception("The system_packages parameters cannot contain newlines.")
    self._systemPackages = packages

  def set_docker_run(self, commands: Union[str, None]):
    if commands is None:
      self._dockerRun = None
      return
    if type(commands) != str:
      raise Exception("The docker_run parameter must be a string.")
    if "\n" in commands or "\r" in commands:
      raise Exception("The docker_run parameter cannot contain newlines.")
    self._dockerRun = commands

  def _set_main_function(self, func: Callable[..., Any]):
    self._deployFunc = func
    if callable(func) and self._deployName == None:
      self.set_name(func.__name__)
    tests = self._getTests()
    if tests == None:
      self._testOutputs = []
    else:
      self._testOutputs = self._getTestOutputs(tests)
    return self

  def _getRequirementsTxt(self):
    if self._pythonPackages:
      return "\n".join(self._pythonPackages)
    else:
      return None

  def _prepEnvironment(self):
    getJson(
        "jupyter/v1/runtimes/prep_environment", {
            "environment": {
                "requirementsTxt": self._getRequirementsTxt(),
                "pythonVersion": self._pythonVersion,
                "systemPackages": self._systemPackages,
                "dockerRun": self._dockerRun,
            }
        })

  def deploy(self):
    printTemplate("runtime-pre-deploy", None, deploymentName=self._deployName)

    status = self._getStatusNotes()
    rtProps, errors = self._getFuncProps()
    if not status.deployable or len(errors) > 0:
      printTemplate("error", None, errorText="Unable to continue because errors are present.")
      return self

    printTemplate("runtime-deploying",
                  None,
                  deploymentName=self._deployName,
                  warningsList=status.warnings,
                  tipsList=status.tips,
                  errorsList=status.errors)

    sourceFile = self._makeSourceFile(rtProps)
    valueFiles = self._makeAndUploadValueFiles(rtProps)
    resp = getJsonOrPrintError(
        "jupyter/v1/runtimes/create", {
            "runtime": {
                "name": self._deployName,
                "type": self.rtType.value,
                "pyState": {
                    "sourceFile": sourceFile.asDict(),
                    "valueFiles": valueFiles,
                    "name": rtProps.name,
                    "module": self._sourceFileName(),
                    "argNames": rtProps.argNames,
                    "argTypes": rtProps.argTypes,
                    "requirementsTxt": self._getRequirementsTxt(),
                    "pythonVersion": self._pythonVersion,
                    "systemPackages": self._systemPackages,
                    "dockerRun": self._dockerRun
                }
            }
        })
    if not isAuthenticated():
      return None
    if not resp:
      printTemplate("error", None, errorText='Error processing request: no response from server.')
    elif resp.error:
      printTemplate("error", None, errorText=f'Error processing request: {resp.error}')
    elif resp.runtimeOverviewUrl:
      printTemplate("runtime-deployed",
                    None,
                    deploymentName=self._deployName,
                    deployMessage=resp.message,
                    deployTimeWords=self._parseTimeFromDeployMessage(resp.message),
                    runtimeOverviewUrl=resp.runtimeOverviewUrl)
    else:
      printTemplate(
          "error",
          None,
          errorText="Unknown error while processing request (server response in unexpected format).")
    return None

  # a bit of a hack for now
  def _parseTimeFromDeployMessage(self, message: Union[str, None]):
    if not message:
      return None
    if "will be ready in" in message:
      return message.split("will be ready in")[1]
    return None

  def _sourceFileName(self):
    return "source"

  def _makeSourceFile(self, pyProps: RuntimePythonProps):

    def addSpacer(strList: List[str]):
      if len(strList) > 0 and strList[-1] != "":
        strList.append("")

    sourceParts: List[str] = ["import modelbit, sys"]

    if pyProps.namespaceFroms:
      for iAs, iModule in pyProps.namespaceFroms.items():
        sourceParts.append(f"from {iModule} import {iAs}")
    if pyProps.namespaceImports:
      for iAs, iModule in pyProps.namespaceImports.items():
        sourceParts.append(f"import {iModule} as {iAs}")
    addSpacer(sourceParts)

    if pyProps.namespaceVars and pyProps.namespaceVarsDesc:
      for nName, _ in pyProps.namespaceVars.items():
        sourceParts.append(f'{nName} = modelbit.load_value("{nName}") # {pyProps.namespaceVarsDesc[nName]}')

    addSpacer(sourceParts)
    if pyProps.namespaceFunctions:
      for _, fSource in pyProps.namespaceFunctions.items():
        sourceParts.append(fSource)
        addSpacer(sourceParts)

    addSpacer(sourceParts)
    if pyProps.source:
      sourceParts.append("# main function")
      sourceParts.append(pyProps.source)

    sourceParts.append("# to run locally via git & terminal")
    sourceParts.append('if __name__ == "__main__":')
    sourceParts.append(f"  print({pyProps.name}(*sys.argv[1:]))")
    return RuntimeFile(f"{self._sourceFileName()}.py", "\n".join(sourceParts))

  def _makeAndUploadValueFiles(self, pyState: RuntimePythonProps):
    valueFiles: Dict[str, Dict[str, str]] = {}  # is a RuntimeDataMap
    if pyState.namespaceVars and pyState.namespaceVarsDesc:
      for nName, nVal in pyState.namespaceVars.items():
        valueFiles[nName] = {"contentHash": uploadRuntimeObject(nVal, nName)}
    return valueFiles

  def _selfError(self, txt: str):
    printTemplate("error", None, errorText=txt)
    return None

  def _describeHtml(self):
    customStyles = {
        "readyStyle": makeCssStyle({"color": COLORS["success"]}),
        "errorStyle": makeCssStyle({"color": COLORS["error"]}),
        "passStyle": makeCssStyle({
            "color": COLORS["success"],
            "font-weight": "bold"
        }),
        "failStyle": makeCssStyle({
            "color": COLORS["error"],
            "font-weight": "bold"
        }),
    }
    status = self._getStatusNotes()
    funcProps, _ = self._getFuncProps()

    propHeaders = [
        TableHeader("Property", TableHeader.LEFT),
        TableHeader("Value", TableHeader.LEFT, isCode=True),
    ]
    propRows: List[List[str]] = []

    funcSig = '(None)'
    if funcProps.name and funcProps.argNames:
      funcSig = f"{funcProps.name}({', '.join(funcProps.argNames)})"
    propRows.append(["Function", funcSig])

    if funcProps.namespaceFunctions and len(funcProps.namespaceFunctions) > 0:
      nsFuncs = "\n".join([k for k, _ in funcProps.namespaceFunctions.items()])
      propRows.append(["Helpers", nsFuncs])

    if funcProps.namespaceVarsDesc and len(funcProps.namespaceVarsDesc) > 0:
      nsVars = "\n".join([f'{k}: {v}' for k, v in funcProps.namespaceVarsDesc.items()])
      propRows.append(["Values", nsVars])

    nsImports: List[str] = []
    if funcProps.namespaceFroms and len(funcProps.namespaceFroms) > 0:
      for k, v in funcProps.namespaceFroms.items():
        nsImports.append(f'from {v} import {k}')
    if funcProps.namespaceImports and len(funcProps.namespaceImports) > 0:
      for k, v in funcProps.namespaceImports.items():
        nsImports.append(f'import {v} as {k}')
    if len(nsImports) > 0:
      propRows.append(["Imports", '\n'.join(nsImports)])

    propRows.append(["Python Version", self._pythonVersion])

    if self._pythonPackages and len(self._pythonPackages) > 0:
      maxDepsShown = 7
      if len(self._pythonPackages) > maxDepsShown:
        deps = "\n".join([d for d in self._pythonPackages[:maxDepsShown]])
        numLeft = len(self._pythonPackages) - maxDepsShown
        deps += f'\n...and {numLeft} more.'
      else:
        deps = "\n".join([d for d in self._pythonPackages])
      propRows.append(["Python packages", deps])

    if self._systemPackages:
      propRows.append(["System packages", ", ".join(self._systemPackages)])

    if self._dockerRun:
      propRows.append(["Docker RUN", self._dockerRun])

    testHeaders = [
        TableHeader("Status", TableHeader.LEFT, isPassFail=True),
        TableHeader("command", TableHeader.LEFT, isCode=True),
        TableHeader("Result", TableHeader.LEFT, isCode=True),
    ]
    testRows: List[List[Union[str, bool]]] = []
    for tOut in self._testOutputs:
      result: List[str] = []
      metadata = tOut[2]
      if "error" in metadata:
        result.append(f'Error: {metadata["error"]}')
      if "expected" in metadata:
        result.append(f'Expected: {metadata["expected"]}')
      if "received" in metadata:
        result.append(f'Received: {metadata["received"]}')
      testRows.append([tOut[0], tOut[1].command, " ".join(result)])

    return renderTemplate("runtime-description",
                          styles=customStyles,
                          runtimeName=self._deployName,
                          deployable=status.deployable,
                          warningsList=status.warnings,
                          tipsList=status.tips,
                          errorsList=status.errors,
                          properties={
                              "headers": propHeaders,
                              "rows": propRows
                          },
                          tests={
                              "headers": testHeaders,
                              "rows": testRows
                          })

  def _getFuncProps(self):
    errors: List[GenericError] = []
    props: RuntimePythonProps = RuntimePythonProps()
    if not callable(self._deployFunc):
      errors.append(GenericError('The main_function parameter does not appear to be a function.'))
    else:
      props.name = self._deployFunc.__name__
      props.source = self.getFuncSource()
      props.argNames = self.getFuncArgNames()
      props.argTypes = self._annotationsToTypeStr(self._deployFunc.__annotations__)
      nsCollection = NamespaceCollection()
      self._collectNamespaceDeps(self._deployFunc, nsCollection)
      props.namespaceFunctions = nsCollection.functions
      props.namespaceVars = nsCollection.vars
      props.namespaceVarsDesc = self._strValues(nsCollection.vars)
      props.namespaceImports = nsCollection.imports
      props.namespaceFroms = nsCollection.froms
      props.namespaceModules = list(set(nsCollection.allModules))
    return (props, errors)

  def getFuncSource(self):
    if self._sourceOverride:
      return self._sourceOverride
    if not callable(self._deployFunc):
      return None
    return unindent(inspect.getsource(self._deployFunc))

  def getFuncArgNames(self):
    argSpec = inspect.getfullargspec(self._deployFunc)
    if argSpec.varargs:
      return ['...']
    if argSpec.args:
      return argSpec.args
    noArgs: List[str] = []
    return noArgs

  def _annotationsToTypeStr(self, annotations: Dict[str, Any]):
    annoStrs: Dict[str, str] = {}
    for name, tClass in annotations.items():
      try:
        if tClass == Any:
          annoStrs[name] = "Any"
        else:
          annoStrs[name] = tClass.__name__
      except:
        pass
    return annoStrs

  def _collectNamespaceDeps(self, func: Callable[..., Any], collection: NamespaceCollection):
    if not callable(func):
      return collection
    globalsDict = func.__globals__  # type: ignore
    allNames = func.__code__.co_names + func.__code__.co_freevars
    for maybeFuncVarName in allNames:
      if maybeFuncVarName in globalsDict:
        maybeFuncVar = globalsDict[maybeFuncVarName]
        if "__module__" in dir(maybeFuncVar):
          if maybeFuncVar.__module__ == "__main__":  # the user's functions
            argNames = list(maybeFuncVar.__code__.co_varnames or [])
            funcSig = f"{maybeFuncVar.__name__}({', '.join(argNames)})"
            if funcSig not in collection.functions:
              collection.functions[funcSig] = inspect.getsource(maybeFuncVar)
              self._collectNamespaceDeps(maybeFuncVar, collection)
          else:  # functions imported by the user from elsewhere
            if inspect.isclass(maybeFuncVar):
              collection.froms[maybeFuncVarName] = maybeFuncVar.__module__  #
              collection.allModules.append(maybeFuncVar.__module__)
            elif callable(maybeFuncVar) and not self._hasState(maybeFuncVar):
              collection.froms[maybeFuncVarName] = maybeFuncVar.__module__  #
              collection.allModules.append(maybeFuncVar.__module__)
            elif isinstance(maybeFuncVar, object):
              collection.froms[maybeFuncVar.__class__.__name__] = maybeFuncVar.__module__
              collection.allModules.append(maybeFuncVar.__module__)
              collection.vars[maybeFuncVarName] = maybeFuncVar
            else:
              collection.froms[maybeFuncVarName] = f"NYI: {maybeFuncVar.__module__}"
        elif str(maybeFuncVar).startswith('<module'):
          collection.imports[maybeFuncVarName] = maybeFuncVar.__name__
          collection.allModules.append(maybeFuncVar.__name__)
        else:
          collection.vars[maybeFuncVarName] = maybeFuncVar

  def _hasState(self, obj: Any) -> bool:
    try:
      return len(obj.__dict__) > 0
    except:
      return False

  def _getStatusNotes(self):
    tips: List[WarningErrorTip] = []
    warnings: List[WarningErrorTip] = []
    errors: List[WarningErrorTip] = []
    rtPyProps: Union[RuntimePythonProps, None] = None

    # Errors
    if not self._deployName:
      errors.append(GenericError("This deployment needs a name."))
    if not self._deployFunc:
      errors.append(GenericError("This deployment needs a deploy_function."))
    else:
      rtPyProps, propErrors = self._getFuncProps()
      if len(propErrors) > 0:
        errors.extend(errors)
    for tOut in self._testOutputs:
      if tOut[0] == False:  # failed test
        errors.append(GenericError("This deployment has failing tests."))
        break
    if not isAuthenticated():
      errors.append(GenericError("You are not logged in to Modelbit. Please log in, then deploy."))

    # Tips
    if len(self._testOutputs) == 0:
      tips.append(
          GenericTip("Add unit tests to improve this deployment.",
                     "https://doc.modelbit.com/#/deployments/unit-tests"))

    # Warnings
    warnings += getMissingPackageWarningsFromEnvironment(getDefaultPythonPackages())
    if (rtPyProps is not None):
      depPackages = self._pythonPackages
      if depPackages is None:
        depPackages = getDefaultPythonPackages()
      warnings += getMissingPackageWarningsFromImportedModules(rtPyProps.namespaceModules, depPackages)

    localPyVersion = getInstalledPythonVersion()
    if self._pythonVersion != localPyVersion:
      warnings.append(DifferentPythonVerWarning(self._pythonVersion, localPyVersion))

    return RuntimeStatusNotes(tips, warnings, errors)

  def _strValues(self, args: Dict[str, Any]):
    newDict: Dict[str, str] = {}
    for k, v in args.items():
      strVal = re.sub(r'\s+', " ", str(v))
      if len(strVal) > 200:
        strVal = strVal[0:200] + "..."
      newDict[k] = strVal
    return newDict

  def _getTests(self):
    if not callable(self._deployFunc):
      return None
    funcName = self._deployFunc.__name__
    funcSource = self.getFuncSource()
    resp = getJsonOrPrintError("jupyter/v1/runtimes/parse_tests", {
        "funcName": funcName,
        "funcSource": funcSource,
    })
    if resp:
      return resp.tests
    return None

  def _runTest(self, test: DeploymentTestDef):
    if not callable(self._deployFunc):
      return False
    if test.args == None:
      return self._deployFunc()
    return self._deployFunc(*test.args)  # type: ignore

  def _getTestOutputs(self,
                      tests: List[DeploymentTestDef]) -> List[Tuple[bool, DeploymentTestDef, Dict[str, str]]]:
    return [self._getTestOutput(t) for t in tests]

  def _getTestOutput(self, test: DeploymentTestDef) -> Tuple[bool, DeploymentTestDef, Dict[str, str]]:

    if test.error:
      errorMessage = "Unable to run test."
      if test.error == DeploymentTestError.UnknownFormat.value:
        errorMessage = "Unknown test format"
      if test.error == DeploymentTestError.ExpectedNotJson.value:
        errorMessage = "Expected value must be JSON serializeable"
      if test.error == DeploymentTestError.CannotParseArgs.value:
        errorMessage = "Unable to parse function arguments."
      return (False, test, {"error": errorMessage})

    result = self._runTest(test)
    try:
      jResult = json.loads(json.dumps(result))  # dump+load helps clean up numbers like 2.00000000001
      if jResult != test.expectedOutput:
        return (False, test, {"expected": json.dumps(test.expectedOutput), "received": json.dumps(jResult)})
    except:
      return (False, test, {"error": "Output must be JSON serializeable.", "received": type(result).__name__})
    return (True, test, {"expected": json.dumps(test.expectedOutput)})


class Deployment(Runtime):
  _runtimeTypeName = "deployment"

  def __init__(self,
               name: Union[str, None] = None,
               deploy_function: Union[Callable[..., Any], None] = None,
               python_version: Union[str, None] = None,
               requirements_txt_filepath: Union[str, None] = None,
               requirements_txt_contents: Union[List[str], None] = None,
               python_packages: Union[List[str], None] = None,
               system_packages: Union[List[str], None] = None,
               docker_run: Union[str, None] = None,
               source_override: Union[str, None] = None):
    Runtime.__init__(self,
                     RuntimeType.Deployment,
                     name=name,
                     main_function=deploy_function,
                     python_version=python_version,
                     requirements_txt_filepath=requirements_txt_filepath,
                     requirements_txt_contents=requirements_txt_contents,
                     python_packages=python_packages,
                     system_packages=system_packages,
                     docker_run=docker_run,
                     source_override=source_override)

  def set_deploy_function(self, func: Callable[..., Any]):
    self._set_main_function(func)



from setuptools import setup  # type: ignore

setup(
    name='modelbit',
    version='0.12.3',
    description='Python package to connect Jupyter notebooks to Modelbit',
    url='https://www.modelbit.com',
    author='Modelbit',
    author_email='tom@modelbit.com',
    license='MIT',
    packages=['modelbit', 'pyaes'],
    package_data={"modelbit": ["*.pyi", "*.png", "templates/*.j2"]},
    # Note: Keep these deps in sync with snowpark config
    install_requires=[
        'pycryptodomex',
        'pandas',
        'tqdm',
        'requests',
        'types-requests',
        'pyyaml',
        'types-PyYAML',
        'jinja2',
        'types-pkg-resources',
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Framework :: IPython',
        'Framework :: Jupyter',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: MacOS',
        'Operating System :: POSIX :: Linux',
        'Operating System :: Microsoft',
        'Programming Language :: Python :: 3',
    ])



from typing import List, Dict, Union

from .utils import timeago
from .helpers import RuntimeInfo, getJsonOrPrintError, isAuthenticated
from .ux import TableHeader, renderTemplate, UserImage


class DeploymentsList:

  def __init__(self):
    self._deployments: List[RuntimeInfo] = []
    resp = getJsonOrPrintError("jupyter/v1/runtimes/list?runtimeType=Deployment")
    if resp and resp.deployments:
      self._deployments = resp.deployments

  def _repr_html_(self):
    if not isAuthenticated():
      return ""
    return self._makeDeploymentsHtmlTable()

  def _makeDeploymentsHtmlTable(self):
    from collections import defaultdict

    if len(self._deployments) == 0:
      return "There are no deployments to show."
    deploymentsByName: Dict[str, List[RuntimeInfo]] = defaultdict(lambda: [])
    for d in self._deployments:
      deploymentsByName[d.name].append(d)

    headers = [
        TableHeader("Name", TableHeader.LEFT, isCode=True),
        TableHeader("Owner", TableHeader.CENTER),
        TableHeader("Version", TableHeader.RIGHT),
        TableHeader("Deployed", TableHeader.LEFT),
    ]
    rows: List[List[Union[str, UserImage]]] = []
    for dList in deploymentsByName.values():
      ld = dList[0]
      connectedAgo = timeago(ld.deployedAtMs)
      rows.append([ld.name, UserImage(ld.ownerInfo.imageUrl, ld.ownerInfo.name), ld.version, connectedAgo])
    return renderTemplate("table", headers=headers, rows=rows)


def list():
  return DeploymentsList()



from typing import Any, Union, Dict, List
import re, inspect, pkg_resources, sys, json
from .secure_storage import getSecureData
from .utils import timeago, sizeOfFmt
from .helpers import ModelPackage, PickledObj, getJsonOrPrintError, isAuthenticated
from .ux import UserImage, TableHeader, printTemplate, renderTemplate


class Model:

  def __init__(self,
               modelObj: Any = None,
               name: Union[str, None] = None,
               properties: Union[Dict[str, Any], None] = None,
               helpers: Union[Dict[str, Any], None] = None,
               modelPackage: Union[ModelPackage, None] = None):
    if modelPackage:
      self._modelPackage = modelPackage
    else:
      self._modelPackage = ModelPackage({})
      self._modelPackage.requirementsTxt = self._getEnvPackages()
      self._modelPackage.pythonVersion = self._getPythonVersion()
      if modelObj:
        self.set_model(modelObj)
      if name:
        self.set_name(name)
      else:
        self._modelPackage.name = self._inferName()
      if properties:
        self.set_properties(properties)
      if helpers:
        self.set_helpers(helpers)

  def set_model(self, modelObj: Any):
    self._addArtifactId(modelObj)
    self._modelPackage.model = PickledObj(obj=modelObj)

  def get_model(self):
    if self._modelPackage.model:
      return self._modelPackage.model.unpickle()
    return None

  def set_name(self, name: str):
    if type(name) == str and re.fullmatch(r"[a-z0-9/\s_-]+", name):
      self._modelPackage.name = name.strip()
    else:
      raise Exception("Model name should be a string of lowercase letters and numbers and spaces")

  def get_name(self):
    return self._modelPackage.name

  def set_helpers(self, helpers: Dict[str, Any]):
    if type(helpers) != dict:
      raise Exception("Helpers must be a dictionary with string keys.")
    self._modelPackage.helpers = {}
    for k, v in helpers.items():
      self.set_helper(k, v)

  def get_helpers(self):
    helpers: Dict[str, Any] = {}
    for hName, hPkl in self._modelPackage.helpers.items():
      helpers[hName] = hPkl.unpickle()
    return helpers

  def set_helper(self, name: str, value: Any):
    if type(name) != str:
      raise Exception("Helper names must be strings.")
    self._modelPackage.helpers[name] = PickledObj(obj=value)

  def get_helper(self, name: str):
    if name in self._modelPackage.helpers:
      return self._modelPackage.helpers[name].unpickle()
    return None

  def set_properties(self, props: Dict[str, Any]):
    self._modelPackage.properties = {}
    for k, v in props.items():
      if type(k) != str:  # type: ignore
        raise Exception("Property names must be strings.")
      self._modelPackage.properties[k] = v

  def get_properties(self):
    return self._modelPackage.properties

  def set_property(self, name: str, val: Any):
    self._modelPackage.properties[name] = val

  def get_property(self, name: str):
    return self._modelPackage.properties.get(name, None)

  def save(self):
    if self._modelPackage.model == None:
      raise Exception("Unable to save because no model has been supplied.")
    printTemplate("message", None, msgText='Saving model...')
    resp = getJsonOrPrintError("jupyter/v1/models/create", {"modelPackage": self._modelPackage.asDict()})
    if not isAuthenticated():
      return None
    if not resp:
      printTemplate("error", None, errorText="Unable to save model: no response from server.")
    elif resp.error:
      printTemplate("error", None, errorText=f'Unable to save model: {resp.error}')
    elif resp.modelOverviewUrl:
      printTemplate("model-saved",
                    None,
                    modelName=self._modelPackage.name,
                    modelOverviewUrl=resp.modelOverviewUrl)
    else:
      printTemplate("error",
                    None,
                    errorText="Unknown error while saving model (server response in unexpected format).")
    return None

  def _repr_html_(self):
    if not isAuthenticated():
      return ""
    headers = [
        TableHeader("Property", TableHeader.LEFT),
        TableHeader("Value", TableHeader.LEFT, isCode=True),
    ]
    rows: List[List[str]] = []
    ms = self._modelPackage
    rows.append(["Description", ms.model.desc if ms.model and ms.model.desc else ""])
    rows.append(["Module", ms.model.kind if ms.model and ms.model.kind else ""])
    rows.append(["Size", sizeOfFmt(ms.model.size) if ms.model and ms.model.size else ""])

    if ms.properties is not None:
      rows.append(["Properties", json.dumps(ms.properties)])
    else:
      rows.append(["Properties", "None"])

    helpers = "\n".join([f'{k}: {v.desc if v.desc else ""}' for k, v in ms.helpers.items()])
    rows.append(["Helpers", helpers if helpers else "None"])

    modelName = ms.name if ms.name else "<unnamed model>"
    return renderTemplate("model-description", modelName=modelName, headers=headers, rows=rows)

  def _getEnvPackages(self):
    pks = sorted(["%s==%s" % (i.key, i.version) for i in pkg_resources.working_set])  # type: ignore
    return "\n".join(pks)

  def _getPythonVersion(self):
    info = sys.version_info
    return f"{info.major}.{info.minor}"

  def _addArtifactId(self, obj: Any):
    try:
      obj.__setattr__("mb_uuid", self._modelPackage.uuid)
    except:
      pass

  def _inferName(self):
    try:
      codeContexts = [f.code_context for f in inspect.stack()]
      for ccList in codeContexts:
        if not ccList:
          continue
        for cc in ccList:
          captures = re.search(r"\.(save_model|Model)\(([^\s,)]+)", cc)
          if captures:
            return captures.group(2)
    except Exception as _:
      pass
    return "unnamed model"


class ModelsList:

  def __init__(self):
    self._modelPackages: List[ModelPackage] = []
    resp = getJsonOrPrintError("jupyter/v1/models/list")
    if resp and resp.models:
      self._modelPackages = resp.models

  def _repr_html_(self):
    if not isAuthenticated():
      return ""
    if len(self._modelPackages) == 0:
      return "There are no models to show."
    headers = [
        TableHeader("Name", TableHeader.LEFT, isCode=True),
        TableHeader("Owner", TableHeader.CENTER),
        TableHeader("Description", TableHeader.LEFT, isCode=True),
        TableHeader("Created", TableHeader.RIGHT),
    ]
    rows: List[List[Union[str, UserImage]]] = []
    for m in self._modelPackages:
      if not m.name or not m.model or not m.ownerInfo or not m.createdAtMs:
        continue

      rows.append([
          m.name,
          UserImage(m.ownerInfo.imageUrl, m.ownerInfo.name),
          m.model.desc if m.model.desc else "",
          timeago(m.createdAtMs),
      ])
    return renderTemplate("table", headers=headers, rows=rows)


def list():
  return ModelsList()


def get(name: str):
  resp = getJsonOrPrintError("jupyter/v1/models/get", {"modelName": name})
  if resp and resp.modelDownloadInfo:
    stStream = getSecureData(resp.modelDownloadInfo, name)
    if not stStream:
      raise Exception("Unable to download model data.")
    jData = json.load(stStream)
    return Model(modelPackage=ModelPackage(jData))



from typing import Dict, Union, Any
import random, os, io
from html.parser import HTMLParser

LIST_TITLES_WARNINGS = ("inconsistency", "inconsistencies")
LIST_TITLES_TIPS = ("tip", "tips")
LIST_TITLES_ERRORS = ("error", "errors")

COLORS = {
    "brand": "#714488",
    "success": "green",
    "error": "#E2548A",
}


class WarningErrorTip:

  def __init__(self, kind: str):
    self.kind = kind


class GenericError(WarningErrorTip):

  def __init__(self, errorText: str):
    super().__init__("GenericError")
    self.errorText = errorText


class MismatchedPackageWarning(WarningErrorTip):

  def __init__(self, desiredPackage: str, similarPackage: str):
    super().__init__("MismatchedPackageWarning")
    self.desiredPackage = desiredPackage
    self.similarPackage = similarPackage


class MissingPackageFromImportWarning(WarningErrorTip):

  def __init__(self, importedModule: str, localPackage: str):
    super().__init__("MissingPackageFromImportWarning")
    self.importedModule = importedModule
    self.localPackage = localPackage


class DifferentPythonVerWarning(WarningErrorTip):

  def __init__(self, desiredVersion: str, localVersion: str):
    super().__init__("DifferentPythonVerWarning")
    self.desiredVersion = desiredVersion
    self.localVersion = localVersion


class GenericTip(WarningErrorTip):

  def __init__(self, tipText: str, docUrl: str):
    super().__init__("GenericTip")
    self.tipText = tipText
    self.docUrl = docUrl


class UserImage:

  def __init__(self, imageUrl: Union[str, None], ownerName: Union[str, None]):
    self.imageUrl = imageUrl
    self.ownerName = ownerName
    if self.imageUrl is None:
      self.imageUrl = "https://app.modelbit.com/images/profile-placeholder.png"


class TableHeader:
  LEFT = 'left'
  CENTER = 'center'
  RIGHT = 'right'

  def __init__(self, name: str, alignment: str = LEFT, isCode: bool = False, isPassFail: bool = False):
    self.name = name
    self.style = makeCssStyle({
        "text-align": alignment,
        "padding": "10px",
        "vertical-align": "middle",
        "line-height": 1,
    })
    self.isCode = isCode
    self.isPassFail = isPassFail
    if alignment not in [self.LEFT, self.CENTER, self.RIGHT]:
      raise Exception(f'Unrecognized alignment: {alignment}')


# From https://stackoverflow.com/questions/753052/strip-html-from-strings-in-python
class MLStripper(HTMLParser):

  def __init__(self):
    super().__init__()
    self.reset()
    self.strict = False
    self.convert_charrefs = True
    self.text = io.StringIO()

  def handle_data(self, data: str):
    self.text.write(data)

  def get_data(self):
    return self.text.getvalue()


def _strip_tags(html: str):
  s = MLStripper()
  s.feed(html)
  return s.get_data()


def _baseElementStyles():
  headerCss: Dict[str, str | int] = {
      "font-weight": "bold",
      "color": COLORS["brand"],
  }
  headerSuccess = headerCss.copy()
  headerSuccess.update({"color": COLORS["success"]})
  headerError = headerCss.copy()
  headerError.update({"color": COLORS["error"]})

  return {
      "base":
          makeCssStyle({}),
      "bottomSpacing":
          makeCssStyle({"margin": "0 0 20px 0"}),
      "container":
          makeCssStyle({"padding": "5px"}),
      "borderedGroup":
          makeCssStyle({
              "padding": "5px",
              "border-left": f"1px solid {COLORS['brand']}",
              "margin-bottom": "10px",
          }),
      "header":
          makeCssStyle(headerCss),
      "headerSuccess":
          makeCssStyle(headerSuccess),
      "headerError":
          makeCssStyle(headerError),
      "errorLabel":
          makeCssStyle({
              "color": COLORS["error"],
              "font-weight": "bold"
          }),
      "link":
          makeCssStyle({
              "color": "#106ba3",
              "text-decoration": "none",
              "cursor": "pointer"
          }),
      "ul":
          makeCssStyle({
              "background-color": "red",
              "padding": "10px",
          }),
      "li":
          makeCssStyle({
              "margin": "0 0 0 10px",
              "list-style": "circle",
          }),
      "code":
          makeCssStyle({
              "font-family": "monospace",
              "color": COLORS["brand"],
              "font-size": "13px",
              "font-weight": "400",
              "background-color": "#f3f4f6",
              "padding": "3px"
          }),
      "codeInTable":
          makeCssStyle({
              "font-family": "monospace",
              "color": COLORS["brand"],
              "font-size": "13px",
              "font-weight": "400",
              "white-space": "pre-wrap",
          }),
      "userImage":
          makeCssStyle({
              "display": "inline-block",
              "border-radius": "9999px",
              "width": "2rem",
              "height": "2rem",
              "background-color": "rgb(229 231 235)"
          })
  }


def makeCssStyle(styles: Dict[str, Union[str, int]]):
  baseStyles = {
      "margin": 0,
      "padding": 0,
      "line-height": 1.75,
      "font-size": "14px",
      "vertical-align": "baseline",
      "list-style": "none",
      "font-family": "Roboto, Arial, sans-serif",
      "background": "none",
  }
  baseStyles.update(styles)
  return "; ".join([f"{s[0]}: {s[1]}" for s in baseStyles.items()]) + ";"


def renderTemplate(templateName: str, styles: Dict[str, str] = {}, **kwargs: Any):
  stylesWithBase = _baseElementStyles()
  stylesWithBase.update(styles)
  from jinja2 import Environment, FileSystemLoader, select_autoescape
  _jinjaTemplates = Environment(loader=FileSystemLoader(
      os.path.join(os.path.dirname(os.path.realpath(__file__)), 'templates')),
                                autoescape=select_autoescape(['html.j2']))
  renderId = f"mb-{random.randint(1, 999999999)}"
  return _jinjaTemplates.get_template(f"{templateName}.html.j2").render(styles=stylesWithBase,
                                                                        renderId=renderId,
                                                                        **kwargs)


def printTemplate(templateName: str, displayId: Union[str, None], styles: Dict[str, str] = {}, **kwargs: Any):
  # TODO: have separate templates for html and txt mode
  txtMode = os.getenv('MB_TXT_MODE')
  try:
    from IPython import display
  except:
    txtMode = "1"

  html = renderTemplate(templateName, styles=styles, **kwargs)
  dispText = _strip_tags(html.replace("<br/>", "\n"))

  if txtMode:
    print(dispText)
  else:
    # Note: Jupyter display clearing doesn't work in hex (or future txt mode), so don't rely on it
    display.display(display.HTML(html), display_id=displayId, clear=bool(displayId))  # type: ignore



from time import sleep
from typing import Union, Any, List, Dict, cast
from enum import Enum
import json, requests, os, uuid
from .utils import deprecated_pickleObj, sizeOfFmt, getEnvOrDefault, deprecated_unpickleObj
from .environment import getInstalledPythonVersion, ALLOWED_PY_VERSIONS, listMissingPackagesFromImports, listMissingPackagesFromPipList
from .ux import COLORS, DifferentPythonVerWarning, MismatchedPackageWarning, MissingPackageFromImportWarning, WarningErrorTip, makeCssStyle, printTemplate
from threading import Thread

pkgVersion: str = ""  # set in __init__
_MAX_DATA_LEN = 50_000_000
_DEFAULT_CLUSTER = "app.modelbit.com"

_cluster = ""
_region = ""
_api_host = ""
_login_host = ""
_api_url = ""
_currentBranch = "main"


def _setUrls(cluster: Union[str, None]):
  global _cluster, _region, _api_host, _login_host, _api_url
  if cluster is None:
    return
  _cluster = cluster
  _region = _cluster.split(".")[0]
  if cluster == "localhost":
    _api_host = f'http://web:3000/'
    _login_host = f'http://localhost:3000/'
  else:
    _api_host = f'https://{_cluster}/'
    _login_host = _api_host
  _api_url = f'{_api_host}api/'


class OwnerInfo:

  def __init__(self, data: Dict[str, Any]):
    self.id: Union[str, None] = data.get("id", None)
    self.name: Union[str, None] = data.get("name", None)
    self.imageUrl: Union[str, None] = data.get("imageUrl", None)


class DatasetDesc:

  def __init__(self, data: Dict[str, Any]):
    self.name: str = data["name"]
    self.sqlModifiedAtMs: Union[int, None] = data.get("sqlModifiedAtMs", None)
    self.query: str = data["query"]
    self.recentResultMs: Union[int, None] = data.get("recentResultMs", None)
    self.numRows: Union[int, None] = data.get("numRows", None)
    self.numBytes: Union[int, None] = data.get("numBytes", None)
    self.ownerInfo = OwnerInfo(data["ownerInfo"])


class ResultDownloadInfo:

  def __init__(self, data: Dict[str, Any]):
    self.id: str = data["id"]
    self.signedDataUrl: str = data["signedDataUrl"]
    self.key64: str = data["key64"]
    self.iv64: str = data["iv64"]


class ObjectUploadInfo:

  def __init__(self, data: Dict[str, Any]):
    self.signedDataUrl: str = data["signedDataUrl"]
    self.key64: str = data["key64"]
    self.iv64: str = data["iv64"]
    self.objectExists: bool = data["objectExists"]


class WhType(Enum):
  Snowflake = 'Snowflake'
  Redshift = 'Redshift'


class GenericWarehouse:

  def __init__(self, data: Dict[str, Any]):
    self.type: WhType = data["type"]
    self.id: str = data["id"]
    self.displayName: str = data["displayName"]
    self.deployStatusPretty: str = data["deployStatusPretty"]
    self.createdAtMs: int = data["createdAtMs"]


class RuntimeFile:

  def __init__(self, name: str, contents: str):
    self.name = name
    self.contents = contents

  def asDict(self):
    return {"name": self.name, "contents": self.contents}


class RuntimePythonProps:
  excludeFromDict: List[str] = ['errors']

  def __init__(self):
    self.source: Union[str, None] = None
    self.name: Union[str, None] = None
    self.argNames: Union[List[str], None] = None
    self.argTypes: Union[Dict[str, str], None] = None
    self.namespaceVarsDesc: Union[Dict[str, str], None] = None
    self.namespaceFunctions: Union[Dict[str, str], None] = None
    self.namespaceImports: Union[Dict[str, str], None] = None
    self.namespaceFroms: Union[Dict[str, str], None] = None
    self.namespaceModules: Union[List[str], None] = None
    self.errors: Union[List[str], None] = None
    self.namespaceVars: Union[Dict[str, Any], None] = None


class RuntimeType(Enum):
  Deployment = 'Deployment'


class RuntimeInfo:

  def __init__(self, data: Dict[str, Any]):
    self.id: str = data["id"]
    self.name: str = data["name"]
    self.version: str = data["version"]
    self.deployedAtMs: int = data["deployedAtMs"]
    self.ownerInfo = OwnerInfo(data["ownerInfo"])


class PickledObj:

  def __init__(self, obj: Any = None, jDict: Union[Dict[str, Any], None] = None):
    self.pkl: Union[str, None] = None
    self.desc: Union[str, None] = None
    self.kind: Union[str, None] = None
    self.size: Union[int, None] = None
    if obj:
      self.pkl = deprecated_pickleObj(obj)
      self.desc = str(obj)
      self.kind = self._getClassName(obj)
      self.size = len(self.pkl)
    elif jDict:
      self.pkl = jDict.get("pkl", None)
      self.desc = jDict.get("desc", None)
      self.kind = jDict.get("kind", None)
      self.size = jDict.get("size", None)

  def unpickle(self):
    if self.pkl:
      return deprecated_unpickleObj(self.pkl)
    else:
      return None

  def asDict(self):
    return self.__dict__

  def _getClassName(self, obj: Any):
    try:
      return f"{obj.__module__}.{obj.__class__.__name__}"
    except:
      return ""


class ModelPackage:

  def __init__(self, data: Dict[str, Any]):
    self.uuid: str = data.get("uuid", str(uuid.uuid4()))
    self.name: Union[str, None] = data.get("name", None)
    self.ownerInfo: Union[OwnerInfo, None] = None
    if "ownerInfo" in data:
      self.ownerInfo = OwnerInfo(data["ownerInfo"])
    self.model: Union[PickledObj, None] = None
    if "model" in data:
      self.model = PickledObj(jDict=data["model"])
    self.helpers: Dict[str, PickledObj] = {}
    if "helpers" in data:
      helpers = cast(Dict[str, Any], data["helpers"])
      for hName, hVal in helpers.items():
        self.helpers[hName] = PickledObj(jDict=hVal)
    self.properties: Dict[str, Any] = data.get("properties", {})
    self.requirementsTxt: Union[str, None] = data.get("requirementsTxt", None)
    self.pythonVersion: Union[str, None] = data.get("pythonVersion", None)
    self.createdAtMs: Union[int, None] = data.get("createdAtMs", None)

  def asDict(self):
    d = self.__dict__.copy()
    if self.model:
      d["model"] = self.model.asDict()
    d["helpers"] = {}
    for hName, hPkl in self.helpers.items():
      d["helpers"][hName] = hPkl.asDict()
    del d["ownerInfo"]
    return d


class RuntimeEnvironment:

  def __init__(self, data: Dict[str, Any]):
    self.pythonVersion: str = data.get("pythonVersion", None)
    self.pythonPackages: Union[List[str], None] = data.get("pythonPackages", None)
    self.systemPackages: Union[List[str], None] = data.get("systemPackages", None)


class DeploymentTestError(Enum):
  UnknownFormat = 'UnknownFormat'
  ExpectedNotJson = 'ExpectedNotJson'
  CannotParseArgs = 'CannotParseArgs'


class DeploymentTestDef:

  def __init__(self, data: Dict[str, Any]):
    self.command: str = data.get("command", "")
    self.expectedOutput: Union[str, Dict[Union[str, int, float, bool], Any]] = data.get("expectedOutput", "")
    self.args: Union[List[Any], None] = data.get("args", None)
    self.error: Union[str, None] = data.get("error", None)


class NotebookEnv:

  def __init__(self, data: Dict[str, Any]):
    self.userEmail: Union[str, None] = data.get("userEmail", None)
    self.signedToken: Union[str, None] = data.get("signedToken")
    self.authenticated: bool = data.get("authenticated", False)
    self.workspaceName: Union[str, None] = data.get("workspaceName", None)
    self.mostRecentVersion: Union[str, None] = data.get("mostRecentVersion", None)
    self.cluster: Union[str, None] = data.get("cluster", None)
    self.defaultEnvironment: Union[RuntimeEnvironment, None] = None
    if "defaultEnvironment" in data and data["defaultEnvironment"] is not None:
      self.defaultEnvironment = RuntimeEnvironment(data["defaultEnvironment"])


class NotebookResponse:

  def __init__(self, data: Dict[str, Any]):
    self.error: Union[str, None] = data.get("error", None)
    self.message: Union[str, None] = data.get("message", None)

    self.notebookEnv: Union[NotebookEnv, None] = None
    if "notebookEnv" in data:
      self.notebookEnv = NotebookEnv(data["notebookEnv"])

    self.datasets: Union[List[DatasetDesc], None] = None
    if "datasets" in data:
      self.datasets = [DatasetDesc(d) for d in data["datasets"]]

    self.dsrDownloadInfo: Union[ResultDownloadInfo, None] = None
    if "dsrDownloadInfo" in data:
      self.dsrDownloadInfo = ResultDownloadInfo(data["dsrDownloadInfo"])

    self.warehouses: Union[List[GenericWarehouse], None] = None
    if "warehouses" in data:
      self.warehouses = [GenericWarehouse(w) for w in data["warehouses"]]

    self.runtimeOverviewUrl: Union[str, None] = None
    if "runtimeOverviewUrl" in data:
      self.runtimeOverviewUrl = data["runtimeOverviewUrl"]

    self.deployments: Union[List[RuntimeInfo], None] = None
    if "deployments" in data:
      self.deployments = [RuntimeInfo(d) for d in data["deployments"]]

    self.modelOverviewUrl: Union[str, None] = None
    if "modelOverviewUrl" in data:
      self.modelOverviewUrl = data["modelOverviewUrl"]

    self.models: Union[List[ModelPackage], None] = None
    if "models" in data:
      self.models = [ModelPackage(m) for m in data["models"]]

    self.modelDownloadInfo: Union[ResultDownloadInfo, None] = None
    if "modelDownloadInfo" in data:
      self.modelDownloadInfo = ResultDownloadInfo(data["modelDownloadInfo"])

    self.tests: Union[List[DeploymentTestDef], None] = None
    if "tests" in data:
      self.tests = [DeploymentTestDef(d) for d in data["tests"]]

    self.objectUploadInfo: Union[ObjectUploadInfo, None] = None
    if "objectUploadInfo" in data:
      self.objectUploadInfo = ObjectUploadInfo(data["objectUploadInfo"])


def getJson(path: str, body: Dict[str, Any] = {}) -> NotebookResponse:
  global _state
  requestToken = _state.signedToken
  if requestToken == None:
    requestToken = os.getenv('MB_RUNTIME_TOKEN')
  data: Dict[str, Any] = {"requestToken": requestToken, "version": pkgVersion, "branch": _currentBranch}
  data.update(body)
  dataLen = len(json.dumps(data))
  if (dataLen > _MAX_DATA_LEN):
    return NotebookResponse({
        "error":
            f'API Error: Request is too large. (Request is {sizeOfFmt(dataLen)} Limit is {sizeOfFmt(_MAX_DATA_LEN)})'
    })
  with requests.post(f'{_api_url}{path}', json=data) as url:  # type: ignore
    nbResp = NotebookResponse(url.json())  # type: ignore
    if nbResp.notebookEnv:
      _state = nbResp.notebookEnv
    return nbResp


def getJsonOrPrintError(path: str, body: Dict[str, Any] = {}):
  nbResp = getJson(path, body)
  if not isAuthenticated():
    performLogin()
    return False
  if nbResp.error:
    printTemplate("error", None, errorText=nbResp.error)
    return False
  return nbResp


def refreshAuthentication() -> bool:
  global _state
  nbResp = getJson("jupyter/v1/login")
  if nbResp.error:
    printTemplate("error", None, errorText=nbResp.error)
    return False
  if nbResp.notebookEnv:
    _state = nbResp.notebookEnv
    _setUrls(_state.cluster)
  return isAuthenticated()


def isAuthenticated() -> bool:
  return _state.authenticated


def performLogin(refreshAuth: bool = False, region: Union[str, None] = None):
  if region is not None:
    _setUrls(f"{region}.modelbit.com")
  if (refreshAuth):
    refreshAuthentication()
  if isAuthenticated():
    printAuthenticatedMessage()
    return
  displayId = "mbLogin"
  printLoginMessage(displayId)

  def pollForLoggedIn():
    triesLeft = 150
    while not isAuthenticated() and triesLeft > 0:
      triesLeft -= 1
      sleep(3)
      refreshAuthentication()
    if isAuthenticated():
      printAuthenticatedMessage(displayId)
    else:
      printTemplate("login-timeout", displayId=displayId)

  loginThread = Thread(target=pollForLoggedIn)
  if os.getenv('MODELBIT_CI') != "1":
    loginThread.start()


def _pipUpgradeInfo():
  if os.getenv('MB_RUNTIME_TOKEN'):
    return None  # runtime environments don't get upgraded
  latestVer = _state.mostRecentVersion

  def ver2ints(ver: str):
    return [int(v) for v in ver.split(".")]

  nbVer = pkgVersion
  if latestVer and ver2ints(latestVer) > ver2ints(nbVer):
    return {"installed": nbVer, "latest": latestVer}
  return None


def _environmentConflicts():
  warnings: List[WarningErrorTip] = []

  de = _state.defaultEnvironment
  if de is None:
    return warnings

  installedPythonVer = getInstalledPythonVersion()
  if getInstalledPythonVersion() != de.pythonVersion:
    warnings.append(DifferentPythonVerWarning(de.pythonVersion, installedPythonVer))
  warnings += getMissingPackageWarningsFromEnvironment(de.pythonPackages)
  return warnings


def getMissingPackageWarningsFromEnvironment(pyPackages: Union[List[str], None]):
  warnings: List[WarningErrorTip] = []
  missingPackages = listMissingPackagesFromPipList(pyPackages)
  if len(missingPackages) > 0:
    for mp in missingPackages:
      desiredPackage, similarPackage = mp
      if similarPackage is not None:
        warnings.append(MismatchedPackageWarning(desiredPackage, similarPackage))
  return warnings


def getMissingPackageWarningsFromImportedModules(importedModules: Union[List[str], None],
                                                 pyPackages: Union[List[str], None]):
  warnings: List[WarningErrorTip] = []
  missingPackages = listMissingPackagesFromImports(importedModules, pyPackages)
  for mp in missingPackages:
    importedModule, pipPackageInstalled = mp
    warnings.append(MissingPackageFromImportWarning(importedModule, pipPackageInstalled))
  return warnings


def printAuthenticatedMessage(displayId: Union[str, None] = None):
  inRegion: Union[str, None] = None
  if _cluster != _DEFAULT_CLUSTER:
    inRegion = _region
  styles = {
      "connected": makeCssStyle({
          "color": COLORS["success"],
          "font-weight": "bold",
      }),
      "info": makeCssStyle({
          "font-family": "monospace",
          "font-weight": "bold",
          "color": COLORS["brand"],
      })
  }
  printTemplate("authenticated",
                displayId,
                styles=styles,
                email=_state.userEmail,
                workspace=_state.workspaceName,
                inRegion=inRegion,
                needsUpgrade=_pipUpgradeInfo(),
                warningsList=_environmentConflicts())


def printLoginMessage(displayId: Union[str, None] = None):
  if (_state.signedToken == None or type(_state.signedToken) != str):
    raise Exception("Signed token missing, cannot authenticate.")
  displayUrl = f'modelbit.com/t/{_state.signedToken[0:10]}...'
  linkUrl = f'{_login_host}/t/{_state.signedToken}'
  printTemplate("login",
                displayId=displayId,
                displayUrl=displayUrl,
                linkUrl=linkUrl,
                needsUpgrade=_pipUpgradeInfo())


def _runtimeToken():  # type: ignore
  return _state.signedToken


def setCurrentBranch(branch: str):
  global _currentBranch
  if type(branch) != str:
    raise Exception("Branch must be a string.")
  _currentBranch = branch


def getCurrentBranch():
  return _currentBranch


def getDefaultPythonVersion():
  if _state.defaultEnvironment is None:
    installedVersion = getInstalledPythonVersion()
    if installedVersion in ALLOWED_PY_VERSIONS:
      return installedVersion
    return "3.8"
  else:
    return _state.defaultEnvironment.pythonVersion


def getDefaultPythonPackages() -> Union[List[str], None]:
  if _state.defaultEnvironment is None:
    return []
  return _state.defaultEnvironment.pythonPackages


def getDefaultSystemPackages() -> Union[List[str], None]:
  if _state.defaultEnvironment is None:
    return []
  return _state.defaultEnvironment.systemPackages


# set defaults
_setUrls(getEnvOrDefault("MB_JUPYTER_CLUSTER", _DEFAULT_CLUSTER))
_state = NotebookEnv({})
